REBOL [
	Title: "Uniserve: Async HTTP protocol"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Version: 0.9.1
	Date: 27/08/2004
	History: [
		0.9.1	10/10/2004 {
			o Fixed a bug in HTTP header matching
			o Fixed a bug when handling body responses of unknown size
		}
		0.9.0	09/10/2004	"History starts"
	]
]

do %../libs/headers.r

system/schemes/default/proxy: make system/schemes/default/proxy [
	auth-method: none
]

install-protocol [
	name: 'HTTP
	port-id: 80
	verbose: 0

	nlnl: join newline newline
	
	proto-header: compose [	; headers management
		User-Agent		"Mozilla/4.7 [en] (Win98; I)" ;"Uniserve/HTTP client"
		Accept			"*/*"
		Host			(none)
		Content-Type 	(none)
		Content-Length	(none)
	]
	
	reset-stop: does [stop-at: newline]
	
	make-http-session: does [
		context [
			in: context [
				headers: make block! 10
				got-header?: no
				content:
				status-line:
				code: none
				expected: trigger: 0
			]
			out: context [
				method: 'GET
				headers: copy proto-header
				url: content: none
				request: make string! 1024
			]
			auth: context [
				type: user: pass: state: port: header: save-events: none
			]
			state: none
		 ]
	]
	
	clear-session: func [port /local s list][
		s: port/user-data/in
		s/got-header?: no
		s/content: s/status-line: s/code: none
		s/expected: 0
		s/trigger:  4000
		list: s/headers
		forall list [change at list 2 none list: next list]
		s: port/user-data/out
		s/method: 'GET
		s/headers: copy proto-header
		s/url: s/content: none
		clear s/request
	]
	
	set-stream-size: func [n [integer!]][
		server/user-data/in/trigger: n
	]
	
	body-done: func [data /local su][	
		su: server/user-data
		if su/in/content [on-content server su/in/content]
		on-response server su/in
		su/state: 'done
	]
	
	on-connected: does [
		server/proxy: system/schemes/default/proxy
		reset-stop
		server/user-data: make-http-session
		switch/default server/proxy/auth-method [
			BASIC	[]
			DIGEST	[]
			NTLM	[NTLM-auth-process]
		][
			insert-port server server/target
		]	
	]	

	on-received: func [data /locals len name value pos sin su][	
		su: server/user-data
		sin: su/in
		switch/default su/state [
			header [
				either 2 >= length? data [
					sin/got-header?: yes
					on-header-end server sin
					; TODO: process redirect code / error codes
					su/state: 'body
					either su/out/method = 'HEAD [
						body-done data
					][
						either len: select sin/headers 'Content-length [
							stop-at: sin/expected: to integer! len
							if all [
								positive? sin/trigger
								stop-at >= sin/trigger
							][
								su/state: 'stream
								stop-at: sin/trigger
							]
						][
							su/state: 'body
							stop-at: none
						]
					]	
				][
					either sin/code [
						name: to word! to string! copy/part data pos: find data #":"
						value: trim to string! copy next pos
						trim/with value #"^/"
						trim/with value #"^M"
						h-store sin/headers name value
						on-header-line server name value
					][ 
						either find/part data "HTTP" 4 [
							sin/code: to integer! to string! copy/part next find data #" " 3
							sin/status-line: data: to string! data
							trim/with data #"^/"
							trim/with data #"^M"
							on-status-line server sin/code data
						][		
							; no valid header, switching to raw stream mode !
							su/state: 'stream
							stop-at: none
							sin/expected: 0							
							on-stream-in server data							
						]
					]
				]
			]
			body [
				body-done sin/content: data
			]
			stream [
				on-stream-in server data
				all [
					not zero? sin/expected
					greater? sin/trigger sin/expected: sin/expected - length? data
					zero? sin/trigger: sin/expected
					body-done data
				]
			]
		][
			log/error ["HTTP unknown state:" mold su/state]
		]
		if su/state = 'done [close-server]
	]
	
	on-raw-received: func [data /local sin][
		sin: server/user-data/in
		either server/user-data/state = 'body [
			either sin/content [
				append sin/content data
			][
				sin/content: copy data
			]
		][
			on-received data
		]
	]
	
	on-close-server: has [su][
		if su: server/user-data [
			if all [
				not find [body done] su/state
				not empty? server/locals/in-buffer 
			][
				on-stream-in server server/locals/in-buffer
			]
			if su/state <> 'done [body-done su/in/content]
		]
		reset-stop
	]

	new-insert-port: func [port url [url! string!] /post data [string!] /local out name value proxy?][
		clear-session port
		proxy?: to logic! all [port/proxy/type = 'generic port/proxy/host]
		out: port/user-data/out
		out/url: any [all [proxy? port/url] url]
		
		on-send port out
		
		data: any [all [post data] out/content none]
		repend out/request [out/method #" " out/url " HTTP/1.0" crlf]
		h-store out/headers 'Host any [server/host server/remote-ip]
		h-store out/headers 'Content-Length any [all [data length? data] none]
		if proxy? [h-store out/headers 'Host none]
		
		if port/user-data/auth/header [
			h-store out/headers 
				pick [Proxy-Authorization Authorization] proxy?
				port/user-data/auth/header
		]
		
		foreach-nv-pair out/headers [
			if value [repend out/request [name ": " value crlf]]
		]
		append out/request crlf
		if verbose > 1 [log/info join "request=>^/" mold out/request]
		if data [append out/request data]
		write-server/with out/request port
		port/user-data/state: 'header
		reset-stop	
	]
	
	events: [
		on-status-line		;-- [port [port!] code [integer!] line [string!]]
		on-header-line		;-- [port [port!] name [word!] value [string!]]
		on-header-end		;-- [port [port!] response [object!]]
		on-content			;-- [port [port!] data [string!]]
		on-response			;-- [port [port!] response [object!]]
		on-redirect			;-- to be done !
		on-send				;-- [port [port!] request [object!]]
		on-stream-in		;-- [port [port!] data [string!]]
		on-error			;-- [code [word!] port [port!]]		; port should be the first arg !
	]

]
